@@ -34,14 +34,19 @@
 #include <signal.h>
 #include <pwd.h>
 #include <ctype.h>
+#include <list>
+#include <string>
 
 #include "util.h"
 #include "Toc.h"
 #include "ScsiIf.h"
 #include "CdrDriver.h"
 #include "dao.h"
+#include "port.h"
 #include "Settings.h"
 #include "Cddb.h"
+#include "TempFileManager.h"
+#include "FormatConverter.h"
 
 
 #ifdef UNIXWARE
@@ -53,7 +58,8 @@ extern "C" {
 
 enum Command { UNKNOWN, SHOW_TOC, SHOW_DATA, READ_TEST, SIMULATE, WRITE,
 	       READ_TOC, DISK_INFO, READ_CD, TOC_INFO, TOC_SIZE, BLANK,
-	       SCAN_BUS, UNLOCK, COPY_CD, READ_CDDB, MSINFO, DRIVE_INFO };
+	       SCAN_BUS, UNLOCK, COPY_CD, READ_CDDB, MSINFO, DRIVE_INFO,
+               DISCID };
 
 static const char *PRGNAME = NULL;
 static const char *TOC_FILE = NULL;
@@ -63,6 +69,8 @@ static const char *SOURCE_SCSI_DEVICE = NULL;
 static const char *DATA_FILENAME = NULL;
 static const char *CDDB_SERVER_LIST = "freedb.freedb.org freedb.freedb.org:/~cddb/cddb.cgi uk.freedb.org uk.freedb.org:/~cddb/cddb.cgi cz.freedb.org cz.freedb.org:/~cddb/cddb.cgi";
 static const char *CDDB_LOCAL_DB_DIR = NULL;
+static const char *TMP_FILE_DIR = NULL;
+static int READING_SPEED = -1;
 static int WRITING_SPEED = -1;
 static int EJECT = 0;
 static int SWAP = 0;
@@ -73,6 +81,7 @@ static int SESSION = 1; // session for read-toc/read-cd
 static int FAST_TOC = 0; // toc reading without sub-channel analysis
 static int PAUSE = 1; // pause before writing
 static int READ_RAW = 0; // read raw sectors
+static int MODE2_MIXED = 1;
 static int REMOTE_MODE = 0;
 static int REMOTE_FD = -1;
 static int RELOAD = 0;
@@ -81,6 +90,8 @@ static int PARANOIA_MODE = 3;
 static int ON_THE_FLY = 0;
 static int WRITE_SIMULATE = 0;
 static int SAVE_SETTINGS = 0;
+static int USER_CAPACITY = 0;
+static int FULL_BURN = 0;
 static int CDDB_TIMEOUT = 60;
 static int WITH_CDDB = 0;
 static int TAO_SOURCE = 0;
@@ -89,6 +100,8 @@ static int KEEPIMAGE = 0;
 static int OVERBURN = 0;
 static int BUFFER_UNDER_RUN_PROTECTION = 1;
 static int WRITE_SPEED_CONTROL = 1;
+static bool KEEP = false;
+static bool PRINT_QUERY = false;
 
 static CdrDriver::BlankingMode BLANKING_MODE = CdrDriver::BLANK_MINIMAL;
 static TrackData::SubChannelMode READ_SUBCHAN_MODE = TrackData::SUBCHAN_NONE;
@@ -169,6 +182,19 @@ static void printVersion()
   message(2, "  Paranoia DAE library - (C) Monty");
   message(2, "");
   message(2, "Check http://cdrdao.sourceforge.net/drives.html#dt for current driver tables.");
+
+  std::list<std::string> list;
+  int num = formatConverter.supportedExtensions(list);
+
+  if (num) {
+    std::string msg = "Format converter enabled for extensions:";
+    std::list<std::string>::iterator i = list.begin();
+    for (;i != list.end(); i++) {
+      msg += " ";
+      msg += (*i);
+    }
+    message(3, msg.c_str());
+  }
   message(1, "");
 }
 
@@ -180,28 +206,35 @@ static void printUsage()
     message(0, "\nUsage: %s <command> [options] [toc-file]", PRGNAME);
     message(0,
 "command:\n"
-"  show-toc  - prints out toc and exits\n"
-"  toc-info  - prints out short toc-file summary\n"
-"  toc-size  - prints total number of blocks for toc\n"
-"  read-toc  - create toc file from audio CD\n"
-"  read-cd   - create toc and rip audio data from CD\n"
-"  read-cddb - contact CDDB server and add data as CD-TEXT to toc-file\n"
-"  show-data - prints out audio data and exits\n"
-"  read-test - reads all audio files and exits\n"
-"  disk-info - shows information about inserted medium\n"
-"  msinfo    - shows multi session info, output is suited for scripts\n"
-"  unlock    - unlock drive after failed writing\n"
-"  blank     - blank a CD-RW\n"
-"  scanbus   - scan for devices\n"
-"  simulate  - shortcut for 'write --simulate'\n"
-"  write     - writes CD\n"
-"  copy      - copies CD\n");
+"  show-toc   - prints out toc and exits\n"
+"  toc-info   - prints out short toc-file summary\n"
+"  toc-size   - prints total number of blocks for toc\n"
+"  read-toc   - create toc file from audio CD\n"
+"  read-cd    - create toc and rip audio data from CD\n"
+"  read-cddb  - contact CDDB server and add data as CD-TEXT to toc-file\n"
+"  show-data  - prints out audio data and exits\n"
+"  read-test  - reads all audio files and exits\n"
+"  disk-info  - shows information about inserted medium\n"
+"  discid     - prints out CDDB information\n"
+"  msinfo     - shows multi session info, output is suited for scripts\n"
+"  drive-info - shows drive information\n"
+"  unlock     - unlock drive after failed writing\n"
+"  blank      - blank a CD-RW\n"
+"  scanbus    - scan for devices\n"
+"  simulate   - shortcut for 'write --simulate'\n"
+"  write      - writes CD\n"
+"  copy       - copies CD\n");
     
     message (0, "\n Try '%s <command> -h' to get a list of available options\n", PRGNAME);
     break;
     
   case SHOW_TOC:
-    message(0, "\nUsage: %s show-toc [-v #] toc-file\n", PRGNAME);
+    message(0, "\nUsage: %s show-toc [options] toc-file", PRGNAME);
+    message(0,
+"options:\n"
+"  --tmpdir <path>         - sets directory for temporary wav files\n"
+"  --keep                  - keep generated temp wav files after exit\n"
+"  -v #                    - sets verbose level\n");
     break;
     
   case SHOW_DATA:
@@ -216,17 +249,25 @@ static void printUsage()
     message(0, "\nUsage: %s simulate [options] toc-file", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  --speed <writing-speed> - selects writing speed\n"
-"  --multi                 - session will not be not closed\n"
+"  --multi                 - session will not be closed\n"
 "  --overburn              - allow to overburn a medium\n"
+"  --full-burn             - force burning to the outer disk edge\n"
+"                            with '--driver generic-mmc-raw'\n"
+"  --capacity <minutes>    - sets disk capacity for '--full-burn'\n"
+"                            you must specify this when using blanks bigger\n"
+"                            than 80 min. (90,99,etc.)\n"
+"                            because they seems like 80 min. blanks\n"
 "  --eject                 - ejects cd after simulation\n"
 "  --swap                  - swap byte order of audio files\n"
 "  --buffers #             - sets fifo buffer size (min. 10)\n"
 "  --reload                - reload the disk if necessary for writing\n"
 "  --force                 - force execution of operation\n"
+"  --tmpdir <path>         - sets directory for temporary wav files\n"
+"  --keep                  - keep generated temp wav files after exit\n"
 "  -v #                    - sets verbose level\n"
 "  -n                      - no pause before writing\n",
 	    SCSI_DEVICE);
@@ -236,23 +277,31 @@ static void printUsage()
     message(0, "\nUsage: %s write [options] toc-file", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  --simulate              - just perform a write simulation\n"
 "  --speed <writing-speed> - selects writing speed\n"
-"  --multi                 - session will not be not closed\n"
+"  --multi                 - session will not be closed\n"
 "  --buffer-under-run-protection #\n"
 "                          - 0: disable buffer under run protection\n"
 "                            1: enable buffer under run protection (default)\n"
 "  --write-speed-control # - 0: disable writing speed control by the drive\n"
 "                            1: enable writing speed control (default)\n" 
 "  --overburn              - allow to overburn a medium\n"
+"  --full-burn             - force burning to the outer disk edge\n"
+"                            with '--driver generic-mmc-raw'\n"
+"  --capacity <minutes>    - sets disk capacity for '--full-burn'\n"
+"                            you must specify this when using blanks bigger\n"
+"                            than 80 min. (90,99,etc.)\n"
+"                            because they seems like 80 min. blanks\n"
 "  --eject                 - ejects cd after writing or simulation\n"
 "  --swap                  - swap byte order of audio files\n"
 "  --buffers #             - sets fifo buffer size (min. 10)\n"
 "  --reload                - reload the disk if necessary for writing\n"
 "  --force                 - force execution of operation\n"
+"  --tmpdir <path>         - sets directory for temporary wav files\n"
+"  --keep                  - keep generated temp wav files after exit\n"
 "  -v #                    - sets verbose level\n"
 "  -n                      - no pause before writing\n",
 	    SCSI_DEVICE);
@@ -262,12 +311,14 @@ static void printUsage()
     message(0, "\nUsage: %s read-toc [options] toc-file", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z> - sets SCSI device of CD-ROM reader\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-ROM reader\n"
 "  --driver <id>    - force usage of specified driver for source device\n"
 "  --datafile <filename>   - name of data file placed in toc-file\n"
 "  --session #             - select session\n"
 "  --fast-toc              - do not extract pre-gaps and index marks\n"
 "  --read-raw              - select raw sectors modes for data tracks\n"
+"  --no-mode2-mixed        - don't switch to mode2_mixed\n"
+"  --rspeed <read-speed>   - selects reading speed\n"
 "  --read-subchan <mode>   - defines sub-channel reading mode\n"
 "                            <mode> = rw | rw_raw\n"
 "  --tao-source            - indicate that source CD was written in TAO mode\n"
@@ -285,23 +336,41 @@ static void printUsage()
     message(0, "\nUsage: %s disk-info [options]", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  -v #                    - sets verbose level\n",
 	    SCSI_DEVICE);
     break;
     
+  case DISCID:
+    message(0, "\nUsage: %s discid [options]", PRGNAME);
+    message(0,
+"options:\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
+"                            (default: %s)\n"
+"  --driver <id>           - force usage of specified driver\n"
+"  --cddb-servers <list>   - sets space separated list of CDDB servers\n"
+"  --cddb-timeout #        - timeout in seconds for CDDB server communication\n"
+"  --cddb-directory <path> - path to local CDDB directory where fetched\n"
+"                            CDDB records will be stored\n"
+"  --query-string          - prints out CDDB query only\n"
+"  -v #                    - sets verbose level\n",
+        SCSI_DEVICE);
+    break;
+   
   case READ_CD:
     message(0, "\nUsage: %s read-cd [options] toc-file", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z> - sets SCSI device of CD-ROM reader\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-ROM reader\n"
 "  --driver <id>    - force usage of specified driver for source device\n"
 "  --datafile <filename>   - name of data file placed in toc-file\n"
 "  --session #             - select session\n"
 "  --fast-toc              - do not extract pre-gaps and index marks\n"
 "  --read-raw              - read raw data sectors (including L-EC data)\n"
+"  --no-mode2-mixed        - don't switch to mode2_mixed\n"
+"  --rspeed <read-speed>   - selects reading speed\n"
 "  --read-subchan <mode>   - defines sub-channel reading mode\n"
 "                            <mode> = rw | rw_raw\n"
 "  --tao-source            - indicate that source CD was written in TAO mode\n"
@@ -317,17 +386,28 @@ static void printUsage()
     break;
     
   case TOC_INFO:
-    message(0, "\nUsage: %s toc-info [-v #] toc-file\n", PRGNAME);
+    message(0, "\nUsage: %s toc-info [options] toc-file", PRGNAME);
+    message(0,
+"options:\n"
+"  --tmpdir <path>         - sets directory for temporary wav files\n"
+"  --keep                  - keep generated temp wav files after exit\n"
+"  -v #                    - sets verbose level\n");
     break;
     
   case TOC_SIZE:
-    message(0, "\nUsage: %s toc-size [-v #] toc-file\n", PRGNAME);
+    message(0, "\nUsage: %s toc-size [options] toc-file", PRGNAME);
+    message(0,
+"options:\n"
+"  --tmpdir <path>         - sets directory for temporary wav files\n"
+"  --keep                  - keep generated temp wav files after exit\n"
+"  -v #                    - sets verbose level\n");
+    break;
 
   case BLANK:
     message(0, "\nUsage: %s blank [options]", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  --speed <writing-speed> - selects writing speed\n"
@@ -345,10 +425,22 @@ static void printUsage()
     message(0, "\nUsage: %s unlock [options]", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  --reload                - reload the disk if necessary for writing\n"
+"  --eject                 - ejects cd after unlocking\n"
+"  -v #                    - sets verbose level\n",
+	    SCSI_DEVICE);
+    break;
+    
+  case DRIVE_INFO:
+    message(0, "\nUsage: %s drive-info [options]", PRGNAME);
+    message(0,
+"options:\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
+"                            (default: %s)\n"
+"  --driver <id>           - force usage of specified driver\n"
 "  -v #                    - sets verbose level\n",
 	    SCSI_DEVICE);
     break;
@@ -357,20 +449,27 @@ static void printUsage()
     message(0, "\nUsage: %s copy [options]", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
-"  --source-device <x,y,z> - sets SCSI device of CD-ROM reader\n"
+"  --source-device {<x,y,z>|device} - sets SCSI device of CD-ROM reader\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  --source-driver <id>    - force usage of specified driver for source device\n"
 "  --simulate              - just perform a copy simulation\n"
 "  --speed <writing-speed> - selects writing speed\n"
-"  --multi                 - session will not be not closed\n"
+"  --rspeed <read-speed>   - selects reading speed\n"
+"  --multi                 - session will not be closed\n"
 "  --buffer-under-run-protection #\n"
 "                          - 0: disable buffer under run protection\n"
 "                            1: enable buffer under run protection (default)\n"
 "  --write-speed-control # - 0: disable writing speed control by the drive\n"
 "                            1: enable writing speed control (default)\n" 
 "  --overburn              - allow to overburn a medium\n"
+"  --full-burn             - force burning to the outer disk edge\n"
+"                            with '--driver generic-mmc-raw'\n"
+"  --capacity <minutes>    - sets disk capacity for '--full-burn'\n"
+"                            you must specify this when using blanks bigger\n"
+"                            than 80 min. (90,99,etc.)\n"
+"                            because they seems like 80 min. blanks\n"
 "  --eject                 - ejects cd after writing or simulation\n"
 "  --swap                  - swap byte order of audio files\n"
 "  --on-the-fly            - perform on-the-fly copy, no image file is created\n"
@@ -411,7 +510,7 @@ static void printUsage()
     message(0, "\nUsage: %s msinfo [options]", PRGNAME);
     message(0,
 "options:\n"
-"  --device [path:]<x,y,z>        - sets SCSI device of CD-writer\n"
+"  --device [proto:]{<x,y,z>|device} - sets SCSI device of CD-writer\n"
 "                            (default: %s)\n"
 "  --driver <id>           - force usage of specified driver\n"
 "  --reload                - reload the disk if necessary for writing\n"
@@ -449,6 +548,14 @@ static void importSettings(Command cmd)
 	*ival >= 10) {
       FIFO_BUFFERS = *ival;
     }
+    if ((ival = SETTINGS->getInteger(SET_USER_CAPACITY)) != NULL &&
+	*ival >= 0) {
+      USER_CAPACITY = *ival;
+    }
+    if ((ival = SETTINGS->getInteger(SET_FULL_BURN)) != NULL &&
+	*ival >= 0) {
+      FULL_BURN = *ival;
+    }
   }
 
   if (cmd == READ_CD || cmd == READ_TOC) {
@@ -481,7 +588,8 @@ static void importSettings(Command cmd)
     }
   }
 
-  if (cmd == BLANK || cmd == DISK_INFO || cmd == MSINFO || cmd == UNLOCK) {
+  if (cmd == BLANK || cmd == DISK_INFO || cmd == MSINFO || cmd == UNLOCK ||
+      cmd == DISCID || cmd == DRIVE_INFO) {
     if ((sval = SETTINGS->getString(SET_WRITE_DRIVER)) != NULL) {
       DRIVER_ID = strdupCC(sval);
     }
@@ -492,7 +600,7 @@ static void importSettings(Command cmd)
   }
 
   if (cmd == READ_CDDB || cmd == COPY_CD || cmd == READ_TOC ||
-      cmd == READ_CD) {
+      cmd == READ_CD || cmd == DISCID) {
     if ((sval = SETTINGS->getString(SET_CDDB_SERVER_LIST)) != NULL) {
       CDDB_SERVER_LIST = strdupCC(sval);
     }
@@ -505,7 +613,16 @@ static void importSettings(Command cmd)
 	*ival > 0) {
       CDDB_TIMEOUT = *ival;
     }
+    if ((sval = SETTINGS->getString(SET_TMP_FILE_DIR)) != NULL) {
+        TMP_FILE_DIR = strdupCC(sval);
+    }
   }
+
+  if ((ival = SETTINGS->getInteger(SET_READ_SPEED)) != NULL &&
+      *ival >= 0) {
+    READING_SPEED = *ival;
+  }  
+
 }
 
 static void exportSettings(Command cmd)
@@ -524,6 +641,14 @@ static void exportSettings(Command cmd)
     if (FIFO_BUFFERS > 0) {
       SETTINGS->set(SET_WRITE_BUFFERS, FIFO_BUFFERS);
     }
+
+    if (FULL_BURN > 0) {
+      SETTINGS->set(SET_FULL_BURN, FULL_BURN);
+    }
+
+    if (USER_CAPACITY > 0) {
+      SETTINGS->set(SET_USER_CAPACITY, USER_CAPACITY);
+    }
   }
 
   if (cmd == READ_CD) {
@@ -546,7 +671,8 @@ static void exportSettings(Command cmd)
     SETTINGS->set(SET_READ_PARANOIA_MODE, PARANOIA_MODE);
   }
 
-  if (cmd == BLANK || cmd == DISK_INFO || cmd == MSINFO || cmd == UNLOCK) {
+  if (cmd == BLANK || cmd == DISK_INFO || cmd == MSINFO || cmd == UNLOCK ||
+      cmd == DISCID || cmd == DRIVE_INFO) {
     if (DRIVER_ID != NULL)
       SETTINGS->set(SET_WRITE_DRIVER, DRIVER_ID);
     
@@ -555,7 +681,8 @@ static void exportSettings(Command cmd)
   }
 
   if (cmd == READ_CDDB ||
-      (WITH_CDDB && (cmd == COPY_CD || cmd == READ_TOC || cmd == READ_CD))) {
+      (WITH_CDDB && (cmd == COPY_CD || cmd == READ_TOC || cmd == READ_CD ||
+                     cmd == DISCID))) {
     if (CDDB_SERVER_LIST != NULL) {
       SETTINGS->set(SET_CDDB_SERVER_LIST, CDDB_SERVER_LIST);
     }
@@ -568,6 +695,10 @@ static void exportSettings(Command cmd)
       SETTINGS->set(SET_CDDB_TIMEOUT, CDDB_TIMEOUT);
     }
   }
+
+  if (READING_SPEED >= 0) {
+    SETTINGS->set(SET_READ_SPEED, READING_SPEED);
+  }
 }
 
 static int parseCmdline(int argc, char **argv)
@@ -597,6 +728,9 @@ static int parseCmdline(int argc, char **argv)
   else if (strcmp(*argv, "disk-info") == 0) {
     COMMAND = DISK_INFO;
   }
+  else if (strcmp(*argv, "discid") == 0) {
+    COMMAND = DISCID;
+  }
   else if (strcmp(*argv, "read-cd") == 0) {
     COMMAND = READ_CD;
   }
@@ -694,6 +828,20 @@ static int parseCmdline(int argc, char **argv)
 	  argc--, argv++;
 	}
       }
+      else if (strcmp((*argv) + 2, "rspeed") == 0) {
+        if (argc < 2) {
+          message(-2, "Missing argument after: %s", *argv);
+          return 1;
+        }
+        else {
+          READING_SPEED = atol(argv[1]);
+          if (READING_SPEED < 0) {
+            message(-2, "Illegal reading speed: %s", argv[1]);
+            return 1;
+          }
+          argc--, argv++;
+        }
+      }
       else if (strcmp((*argv) + 2, "speed") == 0) {
 	if (argc < 2) {
 	  message(-2, "Missing argument after: %s", *argv);
@@ -708,6 +856,20 @@ static int parseCmdline(int argc, char **argv)
 	  argc--, argv++;
 	}
       }
+      else if (strcmp((*argv) + 2, "capacity") == 0) {
+	if (argc < 2) {
+	  message(-2, "Missing argument after: %s", *argv);
+	  return 1;
+	}
+	else {
+	  USER_CAPACITY = atol(argv[1]);
+	  if (USER_CAPACITY < 0) {
+	    message(-2, "Illegal disk capacity: %s minutes", argv[1]);
+	    return 1;
+	  }
+	  argc--, argv++;
+	}
+      }
       else if (strcmp((*argv) + 2, "blank-mode") == 0) {
 	if (argc < 2) {
 	  message(-2, "Missing argument after: %s", *argv);
@@ -762,6 +924,9 @@ static int parseCmdline(int argc, char **argv)
       else if (strcmp((*argv) + 2, "swap") == 0) {
 	SWAP = 1;
       }
+      else if (strcmp((*argv) + 2, "query-string") == 0) {
+        PRINT_QUERY = true;
+      }
       else if (strcmp((*argv) + 2, "multi") == 0) {
 	MULTI_SESSION = 1;
       }
@@ -774,12 +939,18 @@ static int parseCmdline(int argc, char **argv)
       else if (strcmp((*argv) + 2, "read-raw") == 0) {
 	READ_RAW = 1;
       }
+      else if (strcmp((*argv) + 2, "no-mode2-mixed") == 0) {
+	MODE2_MIXED = 0;
+      }
       else if (strcmp((*argv) + 2, "reload") == 0) {
 	RELOAD = 1;
       }
       else if (strcmp((*argv) + 2, "force") == 0) {
 	FORCE = 1;
       }
+      else if (strcmp((*argv) + 2, "keep") == 0) {
+	KEEP = true;
+      }
       else if (strcmp((*argv) + 2, "on-the-fly") == 0) {
 	ON_THE_FLY = 1;
       }
@@ -795,6 +966,9 @@ static int parseCmdline(int argc, char **argv)
       else if (strcmp((*argv) + 2, "overburn") == 0) {
 	OVERBURN = 1;
       }
+      else if (strcmp((*argv) + 2, "full-burn") == 0) {
+	FULL_BURN = 1;
+      }
       else if (strcmp((*argv) + 2, "with-cddb") == 0) {
         WITH_CDDB = 1;
       }
@@ -872,6 +1046,16 @@ static int parseCmdline(int argc, char **argv)
 	  argc--, argv++;
 	}
       }
+      else if (strcmp((*argv) + 2, "tmpdir") == 0) {
+          if (argc < 2) {
+              message(-2, "Missing argument after: %s", *argv);
+              return 1;
+          } else {
+              TMP_FILE_DIR = argv[1];
+              SETTINGS->set(SET_TMP_FILE_DIR, TMP_FILE_DIR);
+              argc--, argv++;
+          }
+      }
       else if (strcmp((*argv) + 2, "cddb-timeout") == 0) {
 	if (argc < 2) {
 	  message(-2, "Missing argument after: %s", *argv);
@@ -963,6 +1147,7 @@ static int parseCmdline(int argc, char **argv)
 
   if (COMMAND != DISK_INFO && COMMAND != BLANK && COMMAND != SCAN_BUS &&
       COMMAND != UNLOCK && COMMAND != COPY_CD && COMMAND != MSINFO &&
+      COMMAND != DISCID &&
       COMMAND != DRIVE_INFO) {
     if (argc < 1) {
       message(-2, "Missing toc-file.");
@@ -979,6 +1164,26 @@ static int parseCmdline(int argc, char **argv)
   return 0;
 }
 
+// Commit settings to overall system. Export them.
+static void commitSettings(Settings* SETTINGS, const char* settingsPath)
+{
+  if (TMP_FILE_DIR)
+    tempFileManager.setTempDirectory(TMP_FILE_DIR);
+
+  tempFileManager.setKeepTemps(KEEP);
+
+  if (SAVE_SETTINGS && settingsPath != NULL) {
+    // If we're saving our settings, give up root privileges and
+    // exit. The --save option is only compiled in if setreuid() is
+    // available (because it could be used for a local root exploit).
+    if (giveUpRootPrivileges()) {
+      exportSettings(COMMAND);
+      SETTINGS->write(settingsPath);
+    }
+    exit(0);
+  }
+}
+
 // Selects driver for device of 'scsiIf'.
 static CdrDriver *selectDriver(Command cmd, ScsiIf *scsiIf,
 			       const char *driverId)
@@ -1018,9 +1223,12 @@ static CdrDriver *selectDriver(Command cmd, ScsiIf *scsiIf,
 				   &options);
     // if no driver is selected, yet, try to select a read driver for
     // disk-info
-    if (id == NULL && (cmd == DISK_INFO || cmd == MSINFO))
+    if (id == NULL && (cmd == DISK_INFO || cmd == MSINFO || cmd == DISCID))
       id = CdrDriver::selectDriver(0, scsiIf->vendor(), scsiIf->product(),
 				   &options);
+    // Still no driver, try to autodetect one
+    if (id == NULL)
+      id = CdrDriver::detectDriver(scsiIf, &options);
       
     if (id != NULL) {
       ret = CdrDriver::createDriver(id, options, scsiIf);
@@ -1058,8 +1266,8 @@ static CdrDriver *setupDevice(Command cmd, const char *scsiDevice,
 
   switch (scsiIf->init()) {
   case 1:
-    message(-2, "Please use option '--device [path:]bus,id,lun', e.g. "
-            "--device 0,6,0 or --device ATAPI:0,0,0");
+    message(-2, "Please use option '--device [proto:]bus,id,lun', e.g. "
+            "--device 0,6,0 or --device ATA:0,0,0");
     delete scsiIf;
     return NULL;
     break;
@@ -1093,6 +1301,13 @@ static CdrDriver *setupDevice(Command cmd, const char *scsiDevice,
   while (initTries > 0) {
     // clear unit attention
     cdr->rezeroUnit(0);
+    if (READING_SPEED >= 0) {
+      if (!cdr->rspeed(READING_SPEED)) {
+        message(-2, "Reading speed %d is not supported by device.",
+                READING_SPEED);
+        exit(1);
+      }
+    }
 
     if (checkReady) {
       retry = 0;
@@ -1119,6 +1334,16 @@ static CdrDriver *setupDevice(Command cmd, const char *scsiDevice,
 	
       cdr->rezeroUnit(0);
 
+      if (READING_SPEED >= 0) {
+        message(0, "Setting reading speed %d.",
+                READING_SPEED); 
+        if (cdr->rspeed(READING_SPEED) != 0) {
+          message(-2, "Reading speed %d is not supported by device.",
+                  READING_SPEED);
+          exit(1);
+        }
+      }
+
       if ((di = cdr->diskInfo()) == NULL) {
 	message(-2, "Cannot get disk information.");
 	delete cdr;
@@ -1164,6 +1389,14 @@ static CdrDriver *setupDevice(Command cmd, const char *scsiDevice,
     }
   }
 
+  if (READING_SPEED >= 0) {
+    if (cdr->rspeed(READING_SPEED) != 0) {
+      message(-2, "Reading speed %d is not supported by device.",
+              READING_SPEED);
+      exit(1);
+    }
+  }
+
   return cdr;
 }
 
@@ -1432,7 +1665,14 @@ static int showMultiSessionInfo(DiskInfo *di)
   return 2;
 }
 
-static int readCddb(Toc *toc)
+static void printCddbQuery(Toc *toc)
+{
+  Cddb cddb(toc);
+
+  cddb.printDbQuery();
+}
+
+static int readCddb(Toc *toc, bool showEntry = false)
 {
   int err = 0;
   char *servers = strdupCC(CDDB_SERVER_LIST);
@@ -1555,6 +1795,9 @@ static int readCddb(Toc *toc)
     err = 2; goto fail;
   }
 
+  if (showEntry)
+      cddb.printDbEntry();
+
   if (!cddb.addAsCdText(toc))
     err = 1;
 
@@ -1570,15 +1813,33 @@ static void scanBus()
   int i, len;
   ScsiIf::ScanData *sdata = ScsiIf::scan(&len);
 
-  if (sdata == NULL)
-    return;
-
-  for (i = 0; i < len; i++) {
-      message(0, "%s: %s, %s, %s", sdata[i].dev.c_str(), sdata[i].vendor,
+  if (sdata) {
+    for (i = 0; i < len; i++) {
+      message(0, "%s : %s, %s, %s", sdata[i].dev.c_str(), sdata[i].vendor,
               sdata[i].product, sdata[i].revision);
+    }
+    delete[] sdata;
   }
 
-  delete[] sdata;
+#ifdef SCSI_ATAPI
+  sdata = ScsiIf::scan(&len, "ATA");
+  if (sdata) {
+    for (i = 0; i < len; i++) {
+      message(0, "%-20s %s, %s, %s", sdata[i].dev.c_str(), sdata[i].vendor,
+              sdata[i].product, sdata[i].revision);
+    }
+    delete[] sdata;
+  } else {
+    sdata = ScsiIf::scan(&len, "ATAPI");
+    if (sdata) {
+      for (i = 0; i < len; i++) {
+        message(0, "%-20s %s, %s, %s", sdata[i].dev.c_str(), sdata[i].vendor,
+                sdata[i].product, sdata[i].revision);
+      }
+      delete[] sdata;
+    }
+  }
+#endif
 }
 
 static int checkToc(const Toc *toc)
@@ -1990,27 +2251,35 @@ int main(int argc, char **argv)
     exit(1);
   }
 
+  commitSettings(SETTINGS, settingsPath);
+
   printVersion();
 
-  if (SAVE_SETTINGS && settingsPath != NULL) {
-    exportSettings(COMMAND);
-    SETTINGS->write(settingsPath);
-  }
-
-
   if (COMMAND != READ_TOC && COMMAND != DISK_INFO && COMMAND != READ_CD &&
       COMMAND != BLANK && COMMAND != SCAN_BUS && COMMAND != UNLOCK &&
+      COMMAND != DISCID &&
       COMMAND != COPY_CD && COMMAND != MSINFO && COMMAND != DRIVE_INFO) {
+
+    // Parse TOC file
     toc = Toc::read(TOC_FILE);
 
     if (REMOTE_MODE) {
       unlink(TOC_FILE);
     }
 
-    if (toc == NULL) {
+    // Check and resolve input files paths
+    if (!toc || !toc->resolveFilenames(TOC_FILE)) {
       exitCode = 1; goto fail;
     }
 
+    if (!toc->convertFilesToWav()) {
+      message(-2, "Could not decode audio files from toc file \"%s\".",
+              TOC_FILE);
+      exitCode = 1; goto fail;
+    }
+
+    toc->recomputeLength();
+
     if (COMMAND != SHOW_TOC && COMMAND != READ_CDDB) {
       if (checkToc(toc) != 0) {
 	message(-2, "Toc file \"%s\" is inconsistent.", TOC_FILE);
@@ -2023,12 +2292,14 @@ int main(int argc, char **argv)
   if (COMMAND == SIMULATE || COMMAND == WRITE || COMMAND == READ_TOC ||
       COMMAND == DISK_INFO || COMMAND == READ_CD || COMMAND == BLANK ||
       COMMAND == UNLOCK || COMMAND == COPY_CD || COMMAND == MSINFO ||
+      COMMAND == DISCID ||
       COMMAND == DRIVE_INFO) {
     cdr = setupDevice(COMMAND, SCSI_DEVICE, DRIVER_ID, 
 		      /* init device? */
 		      (COMMAND == UNLOCK) ? 0 : 1,
 		      /* check for ready status? */
-		      (COMMAND == BLANK || COMMAND == DRIVE_INFO) ? 0 : 1,
+              (COMMAND == BLANK || COMMAND == DRIVE_INFO ||
+               COMMAND == DISCID) ? 0 : 1,
 		      /* reset status of medium if not empty? */
 		      (COMMAND == SIMULATE || COMMAND == WRITE) ? 1 : 0,
 		      REMOTE_MODE, RELOAD);
@@ -2045,6 +2316,22 @@ int main(int argc, char **argv)
       exitCode = 1; goto fail;
     }
   }
+  
+  if (COMMAND == SIMULATE || COMMAND == WRITE || COMMAND == COPY_CD) {
+    if (FULL_BURN) {
+      if (DRIVER_ID && strcmp(DRIVER_ID, "generic-mmc-raw") != 0) {
+        message(-2, "You must use the generic-mmc-raw driver to use the "
+                "full-burn option.");
+        exitCode = 1; goto fail;
+      } else {
+        int mins = USER_CAPACITY ? USER_CAPACITY :
+          Msf(cdr->diskInfo()->capacity).min();
+        message(2, "Burning entire %d mins disc.", mins);
+      }
+    }
+    cdr->fullBurn(FULL_BURN);
+    cdr->userCapacity(USER_CAPACITY);
+  }
 
   if (COMMAND == COPY_CD) {
     if (SOURCE_SCSI_DEVICE != NULL) {
@@ -2087,7 +2374,7 @@ int main(int argc, char **argv)
     break;
 
   case DRIVE_INFO:
-    showDriveInfo(cdr->driveInfo(1));
+    showDriveInfo(cdr->driveInfo(true));
     break;
 
   case SHOW_TOC:
@@ -2128,6 +2415,37 @@ int main(int argc, char **argv)
     showDiskInfo(di);
     break;
 
+  case DISCID:
+    if (di->valid.empty && di->empty) {
+      message(-2, "Inserted disk is empty.");
+      exitCode = 1; goto fail;
+    }
+    cdr->subChanReadMode(READ_SUBCHAN_MODE);
+    cdr->rawDataReading(READ_RAW);
+    cdr->mode2Mixed(MODE2_MIXED);
+    cdr->fastTocReading(1);
+    cdr->taoSource(TAO_SOURCE);
+    if (TAO_SOURCE_ADJUST >= 0)
+      cdr->taoSourceAdjust(TAO_SOURCE_ADJUST);
+
+    cdr->force(FORCE);
+
+    if ((toc = cdr->readDiskToc(SESSION,
+                                (DATA_FILENAME == NULL) ?
+                                "data.wav" : DATA_FILENAME)) == NULL) {
+      cdr->rezeroUnit(0);
+      exitCode = 1; goto fail;
+    }
+    else {
+      cdr->rezeroUnit(0);
+
+      if (PRINT_QUERY)
+          printCddbQuery(toc);
+      else
+          readCddb(toc, true);
+    }
+    break;
+   
   case MSINFO:
     switch (showMultiSessionInfo(di)) {
     case 0:
@@ -2162,6 +2480,7 @@ int main(int argc, char **argv)
 
     cdr->subChanReadMode(READ_SUBCHAN_MODE);
     cdr->rawDataReading(READ_RAW);
+    cdr->mode2Mixed(MODE2_MIXED);
     cdr->fastTocReading(FAST_TOC);
     cdr->taoSource(TAO_SOURCE);
     if (TAO_SOURCE_ADJUST >= 0)
@@ -2215,6 +2534,7 @@ int main(int argc, char **argv)
 
     cdr->subChanReadMode(READ_SUBCHAN_MODE);
     cdr->rawDataReading(READ_RAW);
+    cdr->mode2Mixed(MODE2_MIXED);
     cdr->taoSource(TAO_SOURCE);
     if (TAO_SOURCE_ADJUST >= 0)
       cdr->taoSourceAdjust(TAO_SOURCE_ADJUST);
@@ -2234,11 +2554,6 @@ int main(int argc, char **argv)
     else {
       cdr->rezeroUnit(0);
 
-#if defined(HAVE_SETEUID) && defined(HAVE_SETEGID)
-      seteuid(getuid());
-      setegid(getgid());
-#endif
-
       if (WITH_CDDB) {
 	if (readCddb(toc) == 0) {
 	  message(2, "CD-TEXT data was added to toc-file.");
@@ -2484,9 +2799,13 @@ int main(int argc, char **argv)
     message(1, "Trying to unlock drive...");
 
     cdr->abortDao();
+
     if (cdr->preventMediumRemoval(0) != 0) {
       exitCode = 1; goto fail;
     }
+
+    if (EJECT)
+      cdr->loadUnload(1);
     break;
 
   case UNKNOWN:
