@@ -1,7 +1,7 @@
-/* @(#)scsi-linux-sg.c	1.79 04/01/14 Copyright 1997 J. Schilling */
+/* @(#)scsi-linux-sg.c	1.83 04/05/20 Copyright 1997 J. Schilling */
 #ifndef lint
 static	char __sccsid[] =
-	"@(#)scsi-linux-sg.c	1.79 04/01/14 Copyright 1997 J. Schilling";
+	"@(#)scsi-linux-sg.c	1.83 04/05/20 Copyright 1997 J. Schilling";
 #endif
 /*
  *	Interface for Linux generic SCSI implementation (sg).
@@ -65,6 +65,14 @@ static	char __sccsid[] =
 
 #if LINUX_VERSION_CODE >= 0x01031a /* <linux/scsi.h> introduced in 1.3.26 */
 #if LINUX_VERSION_CODE >= 0x020000 /* <scsi/scsi.h> introduced somewhere. */
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
+	#define __KERNEL__
+	#include <asm/types.h>
+	#include <asm/byteorder.h>
+	#undef __KERNEL__
+#endif
+
 /* Need to fine tune the ifdef so we get the transition point right. */
 #include <scsi/scsi.h>
 #else
@@ -94,7 +102,7 @@ static	char __sccsid[] =
  *	Choose your name instead of "schily" and make clear that the version
  *	string is related to a modified source.
  */
-LOCAL	char	_scg_trans_version[] = "scsi-linux-sg.c-1.79";	/* The version for this transport*/
+LOCAL	char	_scg_trans_version[] = "scsi-linux-sg.c-1.83";	/* The version for this transport*/
 
 #ifndef	SCSI_IOCTL_GET_BUS_NUMBER
 #define	SCSI_IOCTL_GET_BUS_NUMBER 0x5386
@@ -136,7 +144,7 @@ LOCAL	char	_scg_trans_version[] = "scsi-linux-sg.c-1.79";	/* The version for thi
  * XXX Should add extra space in buscookies and scgfiles for a "PP bus"
  * XXX and for two or more "ATAPI busses".
  */
-#define	MAX_SCG		16	/* Max # of SCSI controllers */
+#define	MAX_SCG		256	/* Max # of SCSI controllers */
 #define	MAX_TGT		16
 #define	MAX_LUN		8
 
@@ -287,6 +295,8 @@ scgo_help(scgp, f)
 	return (0);
 }
 
+#include <glob.h>
+
 LOCAL int
 scgo_open(scgp, device)
 	SCSI	*scgp;
@@ -301,8 +311,9 @@ scgo_open(scgp, device)
 	register int	t;
 	register int	l;
 	register int	nopen = 0;
-	char		devname[64];
-		BOOL	use_ata = FALSE;
+	char		*devname;
+	BOOL	use_ata = FALSE;
+	glob_t globbuf;
 
 	if (busno >= MAX_SCG || tgt >= MAX_TGT || tlun >= MAX_LUN) {
 		errno = EINVAL;
@@ -328,10 +339,23 @@ scgo_open(scgp, device)
 			 * have a SCSI host bus adapter driver that sends
 			 * the SCSI commands via the ATA hardware. This way,
 			 * the layering model would be honored.
+			 *
+			 * People like Jens Axboe should finally fix the DMA
+			 * bugs in the ide-scsi hostadaptor emulation module
+			 * from Linux instead of publishing childish patches
+			 * to the comment above.
 			 */
 			use_ata = TRUE;
 			device = NULL;
 			if (scgp->overbose) {
+				/*
+				 * I strongly encourage people who believe that
+				 * they need to patch this message away to read
+				 * the messages in the Solaris USCSI libscg
+				 * layer instead of wetting their tissues while
+				 * being unwilling to look besides their
+				 * own belly button.
+				 */
 				js_fprintf((FILE *)scgp->errfile,
 				"Warning: Using badly designed ATAPI via /dev/hd* interface.\n");
 			}
@@ -371,103 +395,98 @@ scgo_open(scgp, device)
 		goto openbydev;
 
 scanopen:
-	if (use_ata) for (i = 0; i <= 25; i++) {
-		js_snprintf(devname, sizeof (devname), "/dev/hd%c", i+'a');
-					/* O_NONBLOCK is dangerous */
-		f = open(devname, O_RDWR | O_NONBLOCK);
-		if (f < 0) {
-			/*
-			 * Set up error string but let us clear it later
-			 * if at least one open succeeded.
-			 */
-			if (scgp->errstr)
-				js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"Cannot open '/dev/hd*'");
-			if (errno != ENOENT && errno != ENXIO && errno != ENODEV) {
+	/*
+	 * Note that it makes no sense to scan less than all /dev/hd* devices
+	 * as even /dev/hda may be a device that talks SCSI (e.g. a ATAPI
+	 * notebook disk or a CD/DVD writer). The CD/DVD writer case may
+	 * look silly but there may be users that did boot from a SCSI hdd
+	 * and connected 4 CD/DVD writers to both IDE cables in the PC.
+	 */
+	if (use_ata) {
+		glob("/dev/hd[a-z]", GLOB_NOSORT, NULL, &globbuf);
+		
+		for (i = 0; globbuf.gl_pathv && globbuf.gl_pathv[i] != NULL ; i++) {
+			devname = globbuf.gl_pathv[i];
+			f = open(devname, O_RDWR | O_NONBLOCK);
+			if (f < 0) {
+				/*
+				 * Set up error string but let us clear it later
+				 * if at least one open succeeded.
+				 */
 				if (scgp->errstr)
 					js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"Cannot open '%s'", devname);
-				return (0);
+						    "Cannot open '/dev/hd*'");
+				if (errno != ENOENT && errno != ENXIO && errno != ENODEV && errno != EACCES) {
+					if (scgp->errstr)
+						js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
+							    "Cannot open '%s'", devname);
+					globfree(&globbuf);
+					return (0);
+				}
+			} else {
+				int	iparm;
+				
+				if (ioctl(f, SG_GET_TIMEOUT, &iparm) < 0) {
+					if (scgp->errstr)
+						js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
+							    "SCSI unsupported with '/dev/hd*'");
+					close(f);
+					continue;
+				}
+				sg_clearnblock(f);	/* Be very proper about this */
+				if (sg_setup(scgp, f, busno, tgt, tlun, devname[7]-'a')) {
+					globfree(&globbuf);
+					return (++nopen);
+				}
+				if (busno < 0 && tgt < 0 && tlun < 0)
+					nopen++;
 			}
-		} else {
-			int	iparm;
-
-			if (ioctl(f, SG_GET_TIMEOUT, &iparm) < 0) {
-				if (scgp->errstr)
-					js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"SCSI unsupported with '/dev/hd*'");
-				close(f);
-				continue;
-			}
-			sg_clearnblock(f);	/* Be very proper about this */
-			if (sg_setup(scgp, f, busno, tgt, tlun, i))
-				return (++nopen);
-			if (busno < 0 && tgt < 0 && tlun < 0)
-				nopen++;
 		}
+		globfree(&globbuf);
 	}
 	if (use_ata && nopen == 0)
 		return (0);
 	if (nopen > 0 && scgp->errstr)
 		scgp->errstr[0] = '\0';
 
-	if (nopen == 0) for (i = 0; i < 32; i++) {
-		js_snprintf(devname, sizeof (devname), "/dev/sg%d", i);
-					/* O_NONBLOCK is dangerous */
-		f = open(devname, O_RDWR | O_NONBLOCK);
-		if (f < 0) {
-			/*
-			 * Set up error string but let us clear it later
-			 * if at least one open succeeded.
-			 */
-			if (scgp->errstr)
-				js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"Cannot open '/dev/sg*'");
-			if (errno != ENOENT && errno != ENXIO && errno != ENODEV) {
+	if (nopen == 0) {
+		glob("/dev/scd[0-9]", GLOB_NOSORT, NULL, &globbuf);
+		glob("/dev/scd[0-9][0-9]", GLOB_NOSORT|GLOB_APPEND, NULL, &globbuf);
+		/*glob("/dev/sg[a-z]", GLOB_NOSORT|GLOB_APPEND, NULL, &globbuf);*/
+		
+		for (i = 0; globbuf.gl_pathv && globbuf.gl_pathv[i] != NULL ; i++) {
+			devname = globbuf.gl_pathv[i];
+
+			f = open(devname, O_RDWR | O_NONBLOCK);
+			if (f < 0) {
+				/*
+				 * Set up error string but let us clear it later
+				 * if at least one open succeeded.
+				 */
 				if (scgp->errstr)
 					js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"Cannot open '%s'", devname);
-				return (0);
+						    "Cannot open '/dev/scd*'");
+				if (errno != ENOENT && errno != ENXIO && errno != ENODEV) {
+					if (scgp->errstr)
+						js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
+							    "Cannot open '%s'", devname);
+					globfree(&globbuf);
+					return (0);
+				}
+			} else {
+				sg_clearnblock(f);	/* Be very proper about this */
+				if (sg_setup(scgp, f, busno, tgt, tlun, -1)) {
+					globfree(&globbuf);
+					return (++nopen);
+				}
+				if (busno < 0 && tgt < 0 && tlun < 0)
+					nopen++;
 			}
-		} else {
-			sg_clearnblock(f);	/* Be very proper about this */
-			if (sg_setup(scgp, f, busno, tgt, tlun, -1))
-				return (++nopen);
-			if (busno < 0 && tgt < 0 && tlun < 0)
-				nopen++;
 		}
 	}
 	if (nopen > 0 && scgp->errstr)
 		scgp->errstr[0] = '\0';
 
-	if (nopen == 0) for (i = 0; i <= 25; i++) {
-		js_snprintf(devname, sizeof (devname), "/dev/sg%c", i+'a');
-					/* O_NONBLOCK is dangerous */
-		f = open(devname, O_RDWR | O_NONBLOCK);
-		if (f < 0) {
-			/*
-			 * Set up error string but let us clear it later
-			 * if at least one open succeeded.
-			 */
-			if (scgp->errstr)
-				js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"Cannot open '/dev/sg*'");
-			if (errno != ENOENT && errno != ENXIO && errno != ENODEV) {
-				if (scgp->errstr)
-					js_snprintf(scgp->errstr, SCSI_ERRSTR_SIZE,
-							"Cannot open '%s'", devname);
-				return (0);
-			}
-		} else {
-			sg_clearnblock(f);	/* Be very proper about this */
-			if (sg_setup(scgp, f, busno, tgt, tlun, -1))
-				return (++nopen);
-			if (busno < 0 && tgt < 0 && tlun < 0)
-				nopen++;
-		}
-	}
-	if (nopen > 0 && scgp->errstr)
-		scgp->errstr[0] = '\0';
 
 openbydev:
 	if (device != NULL && *device != '\0') {
@@ -478,6 +497,15 @@ openbydev:
 				b = -1;
 		}
 		if (scgp->overbose) {
+			/*
+			 * Before you patch this away, are you sure that you
+			 * know what you are going to to?
+			 *
+			 * Note that this is a warning that helps users from
+			 * cdda2wav, mkisofs and other programs (that
+			 * distinguish SCSI addresses from file names) from
+			 * getting unexpected results.
+			 */
 			js_fprintf((FILE *)scgp->errfile,
 			"Warning: Open by 'devname' is unintentional and not supported.\n");
 		}
@@ -1253,7 +1281,7 @@ scgo_send(scgp)
 				if (sp->ux_errno == 0)
 					sp->ux_errno = EIO;
 
-				if (sp->u_sense.cmd_sense != 0 &&
+				if (sp->u_sense.cmd_sense[0] != 0 &&
 				    sp->u_scb.cmd_scb[0] == 0) {
 					/*
 					 * The Linux SCSI system up to 2.4.xx
@@ -1265,6 +1293,21 @@ scgo_send(scgp)
 					sp->u_scb.cmd_scb[0] = ST_CHK_COND;
 					if (sp->sense_count == 0)
 						sp->sense_count = SG_MAX_SENSE;
+
+					if ((sp->u_sense.cmd_sense[2] == 0) &&
+					    (sp->u_sense.cmd_sense[12] == 0) &&
+					    (sp->u_sense.cmd_sense[13] == 0)) {
+						/*
+						 * The Linux SCSI system will
+						 * send a request sense for
+						 * even a dma underrun error.
+						 * Clear CHECK CONDITION state
+						 * in case of No Sense.
+						 */
+						sp->u_scb.cmd_scb[0] = 0;
+						sp->u_sense.cmd_sense[0] = 0;
+						sp->sense_count = 0;
+					}
 				}
 			}
 			break;
@@ -1494,6 +1537,21 @@ sg_rwsend(scgp)
 					status_byte = ST_CHK_COND;
 					if (sgp->hd.sense_len == 0)
 						sgp->hd.sense_len = SG_MAX_SENSE;
+
+					if ((sp->u_sense.cmd_sense[2] == 0) &&
+					    (sp->u_sense.cmd_sense[12] == 0) &&
+					    (sp->u_sense.cmd_sense[13] == 0)) {
+						/*
+						 * The Linux SCSI system will
+						 * send a request sense for
+						 * even a dma underrun error.
+						 * Clear CHECK CONDITION state
+						 * in case of No Sense.
+						 */
+						sp->u_scb.cmd_scb[0] = 0;
+						sp->u_sense.cmd_sense[0] = 0;
+						sp->sense_count = 0;
+					}
 				}
 				break;
 
