@@ -1,6 +1,6 @@
 /*  cdrdao - write audio CD-Rs in disc-at-once mode
  *
- *  Copyright (C) 1998, 1999 Andreas Mueller <mueller@daneb.ping.de>
+ *  Copyright (C) 1998-2001 Andreas Mueller <andreas@daneb.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,38 +16,6 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
-/*
- * $Log: Toc.cc,v $
- * Revision 1.11  1999/04/05 11:03:01  mueller
- * Added CD-TEXT support.
- *
- * Revision 1.10  1999/04/02 20:36:21  mueller
- * Created implementation class that contains all mutual member data.
- *
- * Revision 1.9  1999/03/27 19:52:26  mueller
- * Added data track support.
- *
- * Revision 1.8  1999/01/10 15:10:13  mueller
- * Added functions 'appendTrack()' and 'appendAudioData()'.
- *
- * Revision 1.7  1998/11/15 12:19:13  mueller
- * Added several functions for manipulating track/index marks.
- *
- * Revision 1.6  1998/10/24 14:28:16  mueller
- * Fixed bug in 'readSamples()'.
- *
- * Revision 1.5  1998/10/03 14:32:31  mueller
- * Applied patch from Bjoern Fischer <bfischer@Techfak.Uni-Bielefeld.DE>.
- *
- * Revision 1.4  1998/09/22 19:17:19  mueller
- * Added seeking to and reading of samples for GUI.
- *
- * Revision 1.3  1998/09/06 12:00:26  mueller
- * Used 'message' function for messages.
- *
- */
-
-static char rcsid[] = "$Id: Toc.cc,v 1.11 1999/04/05 11:03:01 mueller Exp mueller $";
 
 #include <config.h>
 
@@ -56,15 +24,25 @@ static char rcsid[] = "$Id: Toc.cc,v 1.11 1999/04/05 11:03:01 mueller Exp muelle
 #include <errno.h>
 #include <string.h>
 #include <ctype.h>
-#include <fstream.h>
+#include <unistd.h>
+
+#include <fstream>
+#include <iostream>
+#include <sstream>
 
 #include "Toc.h"
 #include "util.h"
 #include "TrackDataList.h"
 #include "CdTextItem.h"
+#include "CueParser.h"
+
+#ifdef UNIXWARE
+extern "C" {
+  extern int      strcasecmp(const char *, const char *);
+}
+#endif
 
 extern Toc *parseToc(FILE *fp, const char *filename);
-extern Toc *parseCue(FILE *fp, const char *filename);
 
 Toc::Toc() : length_(0)
 {
@@ -75,6 +53,26 @@ Toc::Toc() : length_(0)
   catalogValid_ = 0;
 }
 
+// copy constructor
+Toc::Toc(const Toc &obj) : length_(0), cdtext_(obj.cdtext_)
+{
+  tocType_ = obj.tocType_;
+
+  if ((catalogValid_ = obj.catalogValid_))
+    memcpy(catalog_, obj.catalog_, 13);
+
+  nofTracks_ = 0;
+  tracks_ = lastTrack_ = NULL;
+
+  // copy all tracks
+  TrackIterator itr(&obj);
+  const Track *trun;
+
+  for (trun = itr.first(); trun != NULL; trun = itr.next())
+    append(trun);
+}
+
+
 Toc::~Toc()
 {
   TrackEntry *run = tracks_;
@@ -152,14 +150,24 @@ Toc *Toc::read(const char *filename)
 
   fclose(fp);
 
-  /*
-  if (ret != NULL)
-    ret->print(cout);
-    */
-
   return ret;
 }
 
+bool Toc::resolveFilenames(const char* filename)
+{
+  // Resolve all relative filenames to absoluate paths wrt to the toc
+  // file current directory.
+  std::string path = filename;
+  path = path.substr(0, path.rfind('/'));
+  if (path.empty()) path = ".";
+
+  for (TrackEntry* t = tracks_; t != NULL; t = t->next)
+    if (!t->track->resolveFilename(path.c_str()))
+      return false;
+
+  return true;
+}
+
 // Writes toc to file with given name.
 // Return: 0: OK
 //         1: error occured
@@ -169,7 +177,7 @@ int Toc::write(const char *filename) const
   assert(filename != NULL);
   assert(*filename != 0);
 
-  ofstream out(filename);
+  std::ofstream out(filename);
 
   if (!out) {
     message(-2, "Cannot open file \"%s\" for writing: %s", filename,
@@ -182,18 +190,42 @@ int Toc::write(const char *filename) const
   return 0;
 }
 
+bool Toc::write(int fd, bool conversions) const
+{
+    assert(fd);
+
+    std::ostringstream oss(std::ostringstream::out);
+    print(oss, conversions);
+
+    std::string ossstr = oss.str();
+    const char* content = ossstr.c_str();
+    int written = ::write(fd, content, strlen(content));
+
+    return (written >= 0);
+}
+
 int Toc::check() const
 {
   TrackEntry *t;
-  int ret;
+  int trackNr;
+  int ret = 0;
 
-  for (t = tracks_; t != NULL; t = t->next) {
-    if ((ret = t->track->check()) != 0) {
-      return ret;
-    }
+  for (t = tracks_, trackNr = 1; t != NULL; t = t->next, trackNr++) {
+    ret |= t->track->check(trackNr);
   }
 
-  return 0;
+  return ret;
+}
+
+bool Toc::recomputeLength()
+{
+  for (TrackEntry* t = tracks_; t; t = t->next) {
+    if (!t->track->recomputeLength())
+      return false;
+  }
+
+  update();
+  return true;
 }
 
 // Sets catalog number. 's' must be a string of 13 digits.
@@ -243,7 +275,7 @@ const char *Toc::catalog() const
 }
 
 // writes contents in TOC file syntax
-void Toc::print(ostream &out) const
+void Toc::print(std::ostream &out, bool conversions) const
 {
   int i;
   TrackEntry *t;
@@ -255,18 +287,37 @@ void Toc::print(ostream &out) const
     for (i = 0; i < 13; i++) {
       out << (char)(catalog(i) + '0');
     }
-    out << "\"" << endl;
+    out << "\"" << std::endl;
   }
 
   cdtext_.print(0, out);
 
   for (t = tracks_, i = 1; t != NULL; t = t->next, i++) {
     out << "\n// Track " << i << "\n";
-    t->track->print(out);
-    out << endl;
+    t->track->print(out, conversions);
+    out << std::endl;
   }
 }
 
+bool Toc::convertFilesToWav()
+{
+  FormatSupport::Status status = formatConverter.convert(this);
+
+  return (status == FormatSupport::FS_SUCCESS);
+}
+
+void Toc::collectFiles(std::set<std::string>& set)
+{
+  for (TrackEntry* t = tracks_; t != NULL; t = t->next)
+    t->track->collectFiles(set);
+}
+
+void Toc::markFileConversion(const char* src, const char* dst)
+{
+  for (TrackEntry* t = tracks_; t != NULL; t = t->next)
+    t->track->markFileConversion(src, dst);
+}
+
 // find track entry that contains given sample number
 // return: found track entry or 'NULL' if sample is out of range
 Toc::TrackEntry *Toc::findTrack(unsigned long sample) const
@@ -324,11 +375,15 @@ int Toc::moveTrackMarker(int trackNr, int indexNr, long lba)
     return 2;
   }
 
-  if (indexNr <= 1 && act->track->type() != TrackData::AUDIO)
+  if (indexNr <= 1 && 
+      (act->track->type() != TrackData::AUDIO ||
+       act->track->subChannelType() != TrackData::SUBCHAN_NONE))
     return 6;
 
   if ((indexNr == 0 || (indexNr == 1 && act->track->start().lba() == 0)) &&
-      act->pred != NULL && act->pred->track->type() != TrackData::AUDIO)
+      act->pred != NULL && 
+      (act->pred->track->type() != TrackData::AUDIO ||
+       act->track->subChannelType() != TrackData::SUBCHAN_NONE))
     return 6;
   
 
@@ -460,10 +515,14 @@ int Toc::removeTrackMarker(int trackNr, int indexNr)
     return 2;
   }
 
-  if (act->track->type() != TrackData::AUDIO && indexNr <= 1)
+  if ((act->track->type() != TrackData::AUDIO ||
+       act->track->subChannelType() != TrackData::SUBCHAN_NONE) &&
+      indexNr <= 1)
     return 3;
 
-  if (act->pred != NULL && act->pred->track->type() != TrackData::AUDIO &&
+  if (act->pred != NULL && 
+      (act->pred->track->type() != TrackData::AUDIO ||
+       act->track->subChannelType() != TrackData::SUBCHAN_NONE) &&
       indexNr <= 1)
     return 3;
 
@@ -552,6 +611,9 @@ int Toc::addTrackMarker(long lba)
   if (act->track->type() != TrackData::AUDIO)
     return 5;
 
+  if (act->track->subChannelType() != TrackData::SUBCHAN_NONE)
+    return 5;
+
   if (lba <= act->start.lba())
     return 2;
   
@@ -565,7 +627,7 @@ int Toc::addTrackMarker(long lba)
   TrackDataList *dataList =
     act->track->removeToEnd(Msf(lba - act->absStart.lba()).samples());
 
-  Track *t = new Track(act->track->type());
+  Track *t = new Track(act->track->type(), act->track->subChannelType());
   t->appendTrackData(dataList);
   delete dataList;
 
@@ -606,11 +668,15 @@ int Toc::addPregap(long lba)
 
   if (act->track->type() != TrackData::AUDIO) 
     return 4;
+
+  if (act->track->subChannelType() != TrackData::SUBCHAN_NONE)
+    return 4;
   
   if (act->next == NULL)
     return 2; // no next track where we could add pre-gap
 
-  if (act->next->track->type() != TrackData::AUDIO) 
+  if (act->next->track->type() != act->track->type() ||
+      act->next->track->subChannelType() != act->track->subChannelType())
     return 4;
   
   if (lba <= act->start.lba())
@@ -635,6 +701,15 @@ int Toc::addPregap(long lba)
   return 0;
 }
 
+void Toc::fixLengths()
+{
+  TrackEntry* te;
+  int i;
+
+  for (i = 0 , te = tracks_; te; te = te->next, i++) {
+    printf("%d : Track %d\n", i, te->trackNr);
+  }
+}
 
 void Toc::checkConsistency()
 {
@@ -661,7 +736,7 @@ void Toc::checkConsistency()
 // first LBA of new track, 'end' is filled with last LBA + 1 of new track.
 void Toc::appendTrack(const TrackDataList *list, long *start, long *end)
 {
-  Track t(TrackData::AUDIO);
+  Track t(TrackData::AUDIO, TrackData::SUBCHAN_NONE);
   const TrackData *run;
 
   for (run = list->first(); run != NULL; run = list->next())
@@ -672,8 +747,7 @@ void Toc::appendTrack(const TrackDataList *list, long *start, long *end)
   unsigned long len = list->length();
 
   if (len < minTime)
-    t.append(SubTrack(SubTrack::DATA,
-		      TrackData(TrackData::AUDIO, minTime - len)));
+    t.append(SubTrack(SubTrack::DATA, TrackData(minTime - len)));
 
   *start = length().lba();
 
@@ -700,6 +774,10 @@ int Toc::appendTrackData(const TrackDataList *list, long *start, long *end)
   if (lastTrack_->track->type() != TrackData::AUDIO)
     return 1;
 
+  if (lastTrack_->track->subChannelType() != TrackData::SUBCHAN_NONE)
+    return 1;
+
+
   *start = length().lba();
 
   for (run = list->first(); run != NULL; run = list->next())
@@ -729,6 +807,9 @@ int Toc::removeTrackData(unsigned long start, unsigned long end,
   if (tent->track->type() != TrackData::AUDIO)
     return 2;
 
+  if (tent->track->subChannelType() != TrackData::SUBCHAN_NONE)
+    return 2;
+
   if (tent != findTrack(end))
     return 1;
 
@@ -751,6 +832,10 @@ int Toc::insertTrackData(unsigned long pos, const TrackDataList *list)
 
   if (tent != NULL && tent->track->type() != TrackData::AUDIO)
     return 1;
+
+  if (tent != NULL && tent->track->subChannelType() != TrackData::SUBCHAN_NONE)
+    return 1;
+
   
   if (tent != NULL) {
     tent->track->insertTrackData(pos - tent->absStart.samples(), list);
@@ -1046,6 +1131,49 @@ int Toc::checkCdTextData() const
   return err;
 }
 
+
+void Toc::trackSummary(int *nofAudioTracks, int *nofMode1Tracks,
+		       int *nofMode2Tracks) const
+{
+  TrackEntry *run;
+
+  if (nofAudioTracks != NULL)
+    *nofAudioTracks = 0;
+
+  if (nofMode1Tracks != NULL)
+    *nofMode1Tracks = 0;
+
+  if (nofMode2Tracks != NULL)
+    *nofMode2Tracks = 0;
+
+  for (run = tracks_; run != NULL; run = run->next) {
+    switch (run->track->type()) {
+    case TrackData::AUDIO:
+      if (nofAudioTracks != NULL)
+	*nofAudioTracks += 1;
+      break;
+
+    case TrackData::MODE1:
+    case TrackData::MODE1_RAW:
+      if (nofMode1Tracks != NULL)
+	*nofMode1Tracks += 1;
+      break;
+
+    case TrackData::MODE2:
+    case TrackData::MODE2_FORM1:
+    case TrackData::MODE2_FORM2:
+    case TrackData::MODE2_FORM_MIX:
+    case TrackData::MODE2_RAW:
+      if (nofMode2Tracks != NULL)
+	*nofMode2Tracks += 1;
+      break;
+      
+    case TrackData::MODE0:
+      break;
+    }
+  }
+}
+
 // Class TrackIterator
 TrackIterator::TrackIterator(const Toc *t)
 {
@@ -1102,6 +1230,13 @@ const Track *TrackIterator::first(Msf &start, Msf &end)
   return next(start, end);
 }
 
+const Track *TrackIterator::first()
+{
+  Msf start, end;
+
+  return first(start, end);
+}
+
 const Track *TrackIterator::next(Msf &start, Msf &end)
 {
   Track *t;
@@ -1118,6 +1253,12 @@ const Track *TrackIterator::next(Msf &start, Msf &end)
   }
 }
 
+const Track *TrackIterator::next()
+{
+  Msf start, end;
+
+  return next(start, end);
+}
 
 // Class TocReader
 
@@ -1187,6 +1328,7 @@ void TocReader::closeData()
   }
 }
 
+#if 0
 long TocReader::readData(long lba, char *buf, long len)
 {
   long n;
@@ -1229,6 +1371,7 @@ long TocReader::readData(long lba, char *buf, long len)
  
   return nread;
 }
+#endif
 
 // seeks to specified sample (absolute position)
 // return: 0: OK
@@ -1306,3 +1449,8 @@ long TocReader::readSamples(Sample *buf, long len)
  
   return nread;
 }
+
+const char* TocReader::curFilename()
+{
+    return reader.curFilename();
+}
