@@ -242,7 +242,12 @@ void Track::update()
 //         1: given start behind track end
 int Track::start(Msf s)
 {
-  if (s.lba() >= length_.lba()) {
+  // A length of zero gets a pass for this test. The track length is
+  // optional for the last track entry of the TOC file, because the
+  // actual length of the track filename is not necessarily known at
+  // this stage (mp3...). The actual length will be filled in later by
+  // recomputeLength().
+  if (s.lba() >= length_.lba() && (length_.lba() != 0)) {
     return 1;
   }
 
@@ -383,7 +388,6 @@ Msf Track::getIndex(int i) const
 
 int Track::check(int trackNr) const
 {
-  SubTrack *st;
   int ret = 0;
 
   if (length().lba() - start().lba() < Msf(0, 4, 0).lba()) {
@@ -391,13 +395,24 @@ int Track::check(int trackNr) const
     ret = 1;
   }
 
-  for (st = subTracks_; st != NULL; st = st->next_) {
+  for (SubTrack* st = subTracks_; st; st = st->next_) {
     ret |= st->check(trackNr);
   }
 
   return ret;
 }
 
+bool Track::recomputeLength()
+{
+  for (SubTrack* st = subTracks_; st; st = st->next_) {
+    if (st->length() == 0) {
+      st->determineLength();
+    }
+  }
+
+  update();
+}
+
 // Sets ISRC code. Expected string: "CCOOOYYSSSSS"
 //                 C: country code (ASCII)
 //                 O: owner code (ASCII)
@@ -490,7 +505,7 @@ int Track::isPadded() const
 }
 
 // writes out track data in TOC file syntax
-void Track::print(std::ostream &out) const
+void Track::print(std::ostream &out, bool conversions) const
 {
   SubTrack *st;
   const char *s;
@@ -533,7 +548,7 @@ void Track::print(std::ostream &out) const
 
 
   for (st = subTracks_; st != NULL; st = st->next_) {
-    st->print(out);
+    st->print(out, conversions);
   }
 
   if (start_.lba() != 0) {
@@ -545,6 +560,44 @@ void Track::print(std::ostream &out) const
   }
 }
 
+void Track::collectFiles(std::set<std::string>& set)
+{
+  SubTrack* st;
+  for (st = subTracks_; st != NULL; st = st->next_) {
+    const char* f = st->filename();
+    if (f)
+      set.insert(f);
+  }
+}
+
+void Track::markFileConversion(const char* src, const char* dst)
+{
+  SubTrack* st;
+  for (st = subTracks_; st != NULL; st = st->next_) {
+    const char* f = st->filename();
+    if (f && strcmp(f, src) == 0) {
+      st->effectiveFilename(dst);
+    }
+  }
+}
+
+bool Track::resolveFilename(const char* path)
+{
+  SubTrack* st;
+  for (st = subTracks_; st != NULL; st = st->next_) {
+    std::string rfilename;
+    const char* f = st->filename();
+    if (f) {
+      if (::resolveFilename(rfilename, f, path)) {
+        st->effectiveFilename(rfilename.c_str());
+      } else {
+        message(-2, "Could not find input file \"%s\".", f);
+        return false;
+      }
+    }
+  }
+}
+
 // Locates 'SubTrack' that contains specified sample.
 // return: found 'SubTrack' or 'NULL' if given sample is out of range
 SubTrack *Track::findSubTrack(unsigned long sample) const
@@ -1894,6 +1947,15 @@ long TrackReader::readSamples(Sample *buf, long len)
   return ret;
 }
 
+const char* TrackReader::curFilename()
+{
+  const TrackData* td = reader.trackData();
+
+  if (td)
+    return td->filename();
+  else
+    return NULL;
+}
 
 SubTrackIterator::SubTrackIterator(const Track *t)
 {
