@@ -114,20 +114,23 @@ static gchar *INDEX_EXTEND_XPM_DATA[] = {
   "........."};
 
 
-SampleDisplay::SampleDisplay()
+SampleDisplay::SampleDisplay():
+	pixmap_(NULL),
+	trackMarkerPixmap_(NULL),
+	indexMarkerPixmap_(NULL),
+	trackMarkerSelectedPixmap_(NULL),
+	indexMarkerSelectedPixmap_(NULL),
+	trackExtendPixmap_(NULL),
+	indexExtendPixmap_(NULL),
+	drawGc_(NULL)
+		
 {
   adjustment_ = new Gtk::Adjustment(0.0, 0.0, 1.0);
-  adjustment_->signal_value_changed().connect(slot(*this,
+  adjustment_->signal_value_changed().connect(mem_fun(*this,
                                                    &SampleDisplay::scrollTo));
 
   trackManager_ = NULL;
 
-  pixmap_ = NULL;
-  trackMarkerPixmap_ = indexMarkerPixmap_ = NULL;
-  trackMarkerSelectedPixmap_ = indexMarkerSelectedPixmap_ = NULL;
-  trackExtendPixmap_ = indexExtendPixmap_ = NULL;
-
-  drawGc_ = NULL;
   width_ = height_ = chanHeight_ = lcenter_ = rcenter_ = 0;
   timeLineHeight_ = timeLineY_ = 0;
   timeTickWidth_ = 0;
@@ -139,40 +142,40 @@ SampleDisplay::SampleDisplay()
   chanSep_ = 10;
 
   cursorControlExtern_ = false;
-  cursorDrawn_ = 0;
+  cursorDrawn_ = false;
   cursorX_ = 0;
 
-  markerSet_ = 0;
+  markerSet_ = false;
   selectionSet_ = false;
-  regionSet_ = 0;
+  regionSet_ = false;
   dragMode_ = DRAG_NONE;
 
   pickedTrackMarker_ = NULL;
   selectedTrack_ = 0;
   selectedIndex_ = 0;
 
-  signal_expose_event().connect(slot(*this,
-                                     &SampleDisplay::handle_expose_event));
+  signal_expose_event().connect(mem_fun(*this,
+                                     &SampleDisplay::handleExposeEvent));
   signal_configure_event().
-    connect(slot(*this, &SampleDisplay::handle_configure_event));
+    connect(mem_fun(*this, &SampleDisplay::handleConfigureEvent));
   signal_motion_notify_event().
-    connect(slot(*this, &SampleDisplay::handle_motion_notify_event));
+    connect(mem_fun(*this, &SampleDisplay::handleMotionNotifyEvent));
   signal_button_press_event().
-    connect(slot(*this, &SampleDisplay::handleButtonPressEvent));
+    connect(mem_fun(*this, &SampleDisplay::handleButtonPressEvent));
   signal_button_release_event().
-    connect(slot(*this,	&SampleDisplay::handleButtonReleaseEvent));
+    connect(mem_fun(*this,	&SampleDisplay::handleButtonReleaseEvent));
   signal_enter_notify_event().
-    connect(slot(*this, &SampleDisplay::handleEnterEvent));
+    connect(mem_fun(*this, &SampleDisplay::handleEnterEvent));
   signal_leave_notify_event().
-    connect(slot(*this, &SampleDisplay::handleLeaveEvent));
+    connect(mem_fun(*this, &SampleDisplay::handleLeaveEvent));
 
-  set_events (Gdk::EXPOSURE_MASK
-	      | Gdk::LEAVE_NOTIFY_MASK
-	      | Gdk::ENTER_NOTIFY_MASK
-	      | Gdk::BUTTON_PRESS_MASK
-	      | Gdk::BUTTON_RELEASE_MASK
-	      | Gdk::POINTER_MOTION_MASK
-	      | Gdk::POINTER_MOTION_HINT_MASK);
+  set_events(Gdk::EXPOSURE_MASK
+             | Gdk::LEAVE_NOTIFY_MASK
+             | Gdk::ENTER_NOTIFY_MASK
+             | Gdk::BUTTON_PRESS_MASK
+             | Gdk::BUTTON_RELEASE_MASK
+             | Gdk::POINTER_MOTION_MASK
+             | Gdk::POINTER_MOTION_HINT_MASK);
 }
 
 void SampleDisplay::setTocEdit(TocEdit *t)
@@ -181,9 +184,9 @@ void SampleDisplay::setTocEdit(TocEdit *t)
 
   Toc *toc = tocEdit_->toc();
 
-  markerSet_ = 0;
+  markerSet_ = false;
   selectionSet_ = false;
-  regionSet_ = 0;
+  regionSet_ = false;
 
   minSample_ = 0;
 
@@ -293,15 +296,15 @@ void SampleDisplay::getView(unsigned long *start, unsigned long *end)
   *end = maxSample_;
 }
 
-int SampleDisplay::getSelection(unsigned long *start, unsigned long *end)
+bool SampleDisplay::getSelection(unsigned long *start, unsigned long *end)
 {
   if (selectionSet_) {
     *start = selectionStartSample_;
     *end = selectionEndSample_;
-    return 1;
+    return true;
   }
   
-  return 0;
+  return false;
 }
 
 int SampleDisplay::getMarker(unsigned long *sample)
@@ -328,16 +331,24 @@ void SampleDisplay::setRegion(unsigned long start, unsigned long end)
   Toc *toc = tocEdit_->toc();
 
   if (end <= start || end >= toc->length().samples()) {
-    regionSet_ = 0;
+    regionSet_ = false;
   }
   else {
     regionStartSample_ = start;
     regionEndSample_ = end;
-    regionSet_ = 1;
+    regionSet_ = true;
   }
 
   setView(minSample_, maxSample_);
+}
 
+void SampleDisplay::clearRegion()
+{
+  bool wasSet = regionSet_;
+  regionSet_ = false;
+  if (wasSet) {
+    setView(minSample_, maxSample_);
+  }
 }
 
 int SampleDisplay::getRegion(unsigned long *start, unsigned long *end)
@@ -457,7 +468,7 @@ gint SampleDisplay::sample2pixel(unsigned long sample)
   return (gint)(sampleStartX_ + val + 0.5);
 }
 
-bool SampleDisplay::handle_configure_event (GdkEventConfigure *event)
+bool SampleDisplay::handleConfigureEvent(GdkEventConfigure *event)
 {
   Glib::RefPtr<Pango::Context> context = get_pango_context();
   Pango::FontMetrics metrics = context->get_metrics(get_style()->get_font());
@@ -513,6 +524,13 @@ bool SampleDisplay::handle_configure_event (GdkEventConfigure *event)
   width_ = get_width();
   height_ = get_height();
 
+  // Don't even try to do anything smart if we haven't received a
+  // reasonable window size yet. This will keep pixmap_ to NULL. This
+  // is important because during startup we don't control how the
+  // configure_event are timed wrt to gcdmaster bringup.
+  if (width_ <= 1 || height_ <= 1)
+    return true;
+
   chanHeight_ = (height_ - timeLineHeight_ - trackLineHeight_ - 2) / 2;
 
   lcenter_ = chanHeight_ / 2 + trackLineHeight_;
@@ -528,13 +546,8 @@ bool SampleDisplay::handle_configure_event (GdkEventConfigure *event)
   sampleEndX_ = width_ - 10;
   sampleWidthX_ = sampleEndX_ - sampleStartX_ + 1;
   
-  // if (pixmap_)
-  // pixmap_->release();
-      
   pixmap_ = Gdk::Pixmap::create(get_window(), get_width(), get_height(), -1);
 
-  //message(0, "handle_configure_event: %d\n", width_);
-
   if (width_ > 100 && height_ > 100)
     updateSamples();
 
@@ -542,7 +555,7 @@ bool SampleDisplay::handle_configure_event (GdkEventConfigure *event)
 }
 
 
-bool SampleDisplay::handle_expose_event (GdkEventExpose *event)
+bool SampleDisplay::handleExposeEvent (GdkEventExpose *event)
 {
   redraw(event->area.x, event->area.y, event->area.width, event->area.height,
 	 0);
@@ -647,7 +660,7 @@ bool SampleDisplay::handleButtonReleaseEvent(GdkEventButton *event)
   return true;
 }
 
-bool SampleDisplay::handle_motion_notify_event (GdkEventMotion *event)
+bool SampleDisplay::handleMotionNotifyEvent (GdkEventMotion *event)
 {
   gint x, y;
   GdkModifierType state;
@@ -754,13 +767,16 @@ bool SampleDisplay::handleLeaveEvent(GdkEventCrossing *event)
 void SampleDisplay::redraw(gint x, gint y, gint width, gint height,
 			   int drawMask)
 {
+  if (pixmap_ == 0)
+    return;
+
   get_window()->draw_drawable(drawGc_, pixmap_, x, y, x, y, width, height);
 
   if ((drawMask & 0x02) == 0)
     drawMarker();
 
   if ((drawMask & 0x01) == 0 && cursorDrawn_) {
-    cursorDrawn_ = 0;
+    cursorDrawn_ = false;
     drawCursor(cursorX_);
   }
 }
@@ -783,7 +799,7 @@ void SampleDisplay::setMarker(unsigned long sample)
     redraw(markerX_, 0, 1, height_, 0x02);
 
   markerSample_ = sample;
-  markerSet_ = 1;
+  markerSet_ = true;
   drawMarker();
 }
 
@@ -792,7 +808,7 @@ void SampleDisplay::clearMarker()
   if (markerSet_)
     redraw(markerX_, 0, 1, height_, 0x02);
 
-  markerSet_ = 0;
+  markerSet_ = false;
 }
 
 
@@ -803,7 +819,7 @@ void SampleDisplay::updateSamples()
 
   Toc *toc = tocEdit_->toc();
 
-  if (!pixmap_)
+  if (pixmap_ == 0)
     return;
 
   gint halfHeight = chanHeight_ / 2;
@@ -851,8 +867,6 @@ void SampleDisplay::updateSamples()
   drawGc_->set_foreground(sampleColor_);
 
   if (bres > 0) {
-    //message(0, "Draw 1");
-
     for (s = minSample_, i = sampleStartX_;
 	 s < maxSample_ && i <= sampleEndX_;
 	 s += res, i++) {
@@ -872,33 +886,27 @@ void SampleDisplay::updateSamples()
 
       pos = double(lnegsum) * halfHeight;
       pos /= SHRT_MAX;
-      //pos /= bres;
       if (pos != 0)
 	pixmap_->draw_line(drawGc_, i, lcenter_, i, lcenter_ - (gint)pos);
       
       pos = double(lpossum) * halfHeight;
       pos /= SHRT_MAX;
-      //pos /= bres;
       if (pos != 0)
 	pixmap_->draw_line(drawGc_, i, lcenter_, i, lcenter_ - (gint)pos);
       
       pos = double(rnegsum) * halfHeight;
       pos /= SHRT_MAX;
-      //pos /= bres;
       if (pos != 0)
 	pixmap_->draw_line(drawGc_, i, rcenter_, i, rcenter_ - (gint)pos);
       
       pos = double(rpossum) * halfHeight;
       pos /= SHRT_MAX;
-      //pos /= bres;
       if (pos != 0)
 	pixmap_->draw_line(drawGc_, i, rcenter_, i, rcenter_ - (gint)pos);
     }
   }
   else if (maxSample_ > 0 && res >= 1) {
 
-    //message(0, "Draw 2: %ld < %ld", res,
-    //        tocEdit_->sampleManager()->blocking());
     TocReader reader(toc);
 
     if (reader.openData() == 0) {
@@ -940,25 +948,21 @@ void SampleDisplay::updateSamples()
 
 	pos = double(lnegsum) * halfHeight;
 	pos /= SHRT_MAX;
-	//pos /= bres;
 	if (pos != 0)
 	  pixmap_->draw_line(drawGc_, i, lcenter_, i, lcenter_ - (gint)pos);
       
 	pos = double(lpossum) * halfHeight;
 	pos /= SHRT_MAX;
-	//pos /= bres;
 	if (pos != 0)
 	  pixmap_->draw_line(drawGc_, i, lcenter_, i, lcenter_ - (gint)pos);
 	
 	pos = double(rnegsum) * halfHeight;
 	pos /= SHRT_MAX;
-	//pos /= bres;
 	if (pos != 0)
 	  pixmap_->draw_line(drawGc_, i, rcenter_, i, rcenter_ - (gint)pos);
 	
 	pos = double(rpossum) * halfHeight;
 	pos /= SHRT_MAX;
-	//pos /= bres;
 	if (pos != 0)
 	  pixmap_->draw_line(drawGc_, i, rcenter_, i, rcenter_ - (gint)pos);
       }
@@ -968,7 +972,6 @@ void SampleDisplay::updateSamples()
     }
   }
   else if (toc != NULL && maxSample_ > minSample_ + 1) {
-    //message(0, "Draw 3");
 
     TocReader reader(toc);
 
@@ -981,11 +984,8 @@ void SampleDisplay::updateSamples()
       gint pos1;
       gint lastPosLeft, lastPosRight;
 
-      //printf("Drawing exact, res=%ld pres=%g %ld-%ld\n", res, pres, minSample_, maxSample_);
-
       if (reader.seekSample(minSample_) == 0 &&
 	  reader.readSamples(sampleBuf, len) == len) {
-	//printf("Drawing exact1\n");
 
 	for (j = 1, di = sampleStartX_ + pres;
 	     j < len && di < sampleEndX_ + 1; j++, di += pres) {
@@ -1083,7 +1083,7 @@ void SampleDisplay::updateSamples()
 
 void SampleDisplay::drawCursor(gint x)
 {
-  if (!pixmap_)
+  if (pixmap_ == 0)
     return;
 
   if (x < sampleStartX_ || x > sampleEndX_)
@@ -1098,7 +1098,7 @@ void SampleDisplay::drawCursor(gint x)
     get_window()->draw_line(drawGc_, x, trackLineY_, x, height_ - 1);
   }
   
-  cursorDrawn_ = 1;
+  cursorDrawn_ = true;
   cursorX_ = x;
 
   if (cursorControlExtern_ == false)
@@ -1109,7 +1109,7 @@ void SampleDisplay::undrawCursor()
 {
   if (cursorDrawn_) {
     redraw(cursorX_, 0, 1, height_, 0x01);
-    cursorDrawn_ = 0;
+    cursorDrawn_ = false;
   }
 }
 
@@ -1117,7 +1117,7 @@ void SampleDisplay::drawTimeTick(gint x, gint y, unsigned long sample)
 {
   char buf[50];
 
-  if (!pixmap_)
+  if (pixmap_ == 0)
     return;
 
   unsigned long min = sample / (60 * 44100);
